dofile_once("data/scripts/lib/utilities.lua")
dofile_once("data/scripts/perks/perk_list.lua")
dofile_once("mods/removable_perks/files/utils.lua")
local perk_id_lookup

function build_unremovable_perks_table()
  if not unremovable_perks then
    unremovable_perks = {}
    for i, action in ipairs(perk_list) do
      if action.func and not action._remove then
        -- print(tostring(action.id ) .. " is unremovable!" .. ("!"):rep(10))
        unremovable_perks[action.id] = true
      end
    end
  end
end

-- Returns a table of custom perk entries ordered by pickup order
function get_active_perks()
  build_unremovable_perks_table()
  if not perk_id_lookup then
    perk_id_lookup = dofile_once("mods/removable_perks/_virtual/perk_id_lookup.lua")
  end
  local perks = {}
  local players = EntityGetWithTag("player_unit")
  if #players > 0 then
    for i, child in ipairs(EntityGetAllChildren(players[1]) or {}) do
      -- Check whether it's a perk or something else (like gokis hero mode etc)
      local ui_icon_component = EntityGetFirstComponentIncludingDisabled(child, "UIIconComponent")
      if ui_icon_component then
        local ui_description = ComponentGetValue2(ui_icon_component, "description")
        local ui_name = ComponentGetValue2(ui_icon_component, "name")
        local icon_sprite_file = ComponentGetValue2(ui_icon_component, "icon_sprite_file")
        local perk_id = perk_id_lookup[ui_name]
        local essence_id = ({
          ["$item_essence_fire"] = "fire",
          ["$item_essence_air"] = "air",
          ["$item_essence_water"] = "water",
          ["$item_essence_laser"] = "laser",
          ["$item_essence_alcohol"] = "alcohol",
        })[ui_name]
        if (perk_id and not unremovable_perks[perk_id]) or essence_id then
          table.insert(perks, {
            id = perk_id or essence_id,
            ui_name = GameTextGetTranslatedOrNot(ui_name),
            ui_description = ui_description,
            ui_icon = icon_sprite_file,
            perk_entity = child,
            is_perk = not essence_id
          })
        end
      end
    end
  end

  return perks
end
