dofile_once("data/scripts/lib/utilities.lua")
local utils = dofile_once("mods/removable_perks/files/utils.lua")
dofile_once("mods/removable_perks/files/get_perks.lua")

if not gui then
  is_open = false
  gui = GuiCreate()
end

function rebuild_perk_table()
  perks = get_active_perks()
  -- local already_encountered = {}
  -- for i, perk in ipairs(perks) do
  --   if already_encountered[perk.id] then
  --     already_encountered[perk.id].count = (already_encountered[perk.id].count or 0) + 1
  --   else
  --     already_encountered[perk.id] = perk
  --   end
  -- end
end

local button_width = 16
local button_height = 16
local screen_w, screen_h = GuiGetScreenDimensions(gui)
GuiStartFrame(gui)
GuiOptionsAdd(gui, GUI_OPTION.NoPositionTween)
if GameGetIsGamepadConnected() then
  GuiOptionsAdd(gui, GUI_OPTION.NonInteractive)
end
local show_button_in_inventory_only = ModSettingGet("removable_perks.show_button_in_inventory_only")
local inventory_gui_comp
local inventory_is_open
if show_button_in_inventory_only then
  local player = EntityGetWithTag("player_unit")[1]
  if player then
    inventory_gui_comp = EntityGetFirstComponentIncludingDisabled(player, "InventoryGuiComponent")
    inventory_is_open = ComponentGetValue2(inventory_gui_comp, "mActive")
    if inventory_is_open and is_open then
      is_open = false
    end
  end
end
if ((show_button_in_inventory_only and (inventory_is_open or is_open)) or not show_button_in_inventory_only) and GuiImageButton(gui, 2, screen_w - 28, 73, "", "mods/removable_perks/files/open_menu_icon.png") then
  is_open = not is_open
  if is_open then
    rebuild_perk_table()
    if show_button_in_inventory_only and inventory_gui_comp then
      ComponentSetValue2(inventory_gui_comp, "mActive", false)
    end
  end
end
if not is_gamepad_active() then
  GuiTooltip(gui, "Remove perks", "Click to open/close a menu that lets you remove perks you have picked up.")
end

local width = 180
local window_x_percent = 50
local window_y_percent = 30

if is_open then
  local animation_speed = 0.1
  if tooltip and tooltip.title then
    -- Perk description shown on hover
    GuiAnimateBegin(gui)
    GuiLayoutBeginVertical(gui, window_x_percent, window_y_percent)
    GuiBeginAutoBox(gui)
    GuiAnimateAlphaFadeIn(gui, 99998, animation_speed, 0.01, false)
    GuiText(gui, -(width/2), -88, tooltip.title)
    GuiColorSetForNextWidget(gui, 1, 1, 1, 0.6)
    GuiText(gui, -(width/2), 0, tooltip.description)
    GuiZSetForNextWidget(gui, 1)
    GuiAnimateScaleIn(gui, 99999, animation_speed, false)
    GuiEndAutoBoxNinePiece(gui, 5)
    GuiLayoutEnd(gui)
    GuiAnimateEnd(gui)
  end
  -- Perk picker window
  GuiAnimateBegin(gui)
  GuiAnimateAlphaFadeIn(gui, 99997, animation_speed, 0.01, false)
  GuiAnimateScaleIn(gui, 99996, animation_speed, false)
  GuiLayoutBeginVertical(gui, window_x_percent, window_y_percent)
  tooltip = tooltip or {}
  GuiBeginAutoBox(gui)
  GuiZSetForNextWidget(gui, -2)
  GuiText(gui, -(width/2), -50, "Removable perks (click to remove):")
  local h = 0
  if was_perk_removed then
    rebuild_perk_table()
  end
  if #perks == 0 then
    GuiZSetForNextWidget(gui, -2)
    local no_perks_found_text = "< No removable perks found >"
    GuiText(gui, -(width/2), 0, no_perks_found_text)
    _, h = GuiGetTextDimensions(gui, no_perks_found_text)
  end
  GuiZSetForNextWidget(gui, -100)
  GuiBeginScrollContainer(gui, 999, -(width/2), 2, width, 105 - h, true, -2, 0)
  was_perk_removed = false
  local something_hovered = false
  for i=1,#perks do
    local x = (i-1) % 10
    local y = math.floor((i-1) / 10)
    GuiZSetForNextWidget(gui, -101)
    local clicked, right_clicked = GuiImageButton(gui, 600+i, x*18, y*18, "", perks[i].ui_icon)
    local _, _, hovered = GuiGetPreviousWidgetInfo(gui)
    something_hovered = something_hovered or hovered
    local drop_perk = not (perks[i].id == "EXTRA_MANA" or perks[i].id == "EXTRA_SLOTS" or perks[i].id == "GAMBLE" or perks[i].id == "MEGA_BEAM_STONE")
    if hovered and last_hovered ~= i then
      last_hovered = i
      -- TODO: Do some line wrapping for long descriptions?
      local title = perks[i].ui_name
      if not drop_perk then
        title = title .. " (Unremovable, REMOVE ICON ONLY)"
      end
      tooltip = { title = title, description = perks[i].ui_description }
    end
    if clicked then
      if perks[i].is_perk then
        utils.remove_perk(perks[i], false, drop_perk)
      else
        utils.remove_essence(perks[i], false, true)
      end
      was_perk_removed = true
      break
    end
  end
  if not something_hovered or was_perk_removed then
    tooltip = {}
    last_hovered = nil
  end
  GuiEndScrollContainer(gui)

  GuiEndAutoBoxNinePiece(gui, 5)
  GuiLayoutEnd(gui)
  GuiAnimateEnd(gui)
end
