dofile_once("data/scripts/lib/utilities.lua")
dofile_once("mods/removable_perks/files/utils.lua")
dofile_once("mods/removable_perks/files/get_perks.lua")
local EZWand = dofile_once("mods/removable_perks/lib/EZWand.lua")

local perk_removal_functions = {
  GOLD_IS_FOREVER = function(entity_id)
    edit_component2(GameGetWorldStateEntity(), "WorldStateComponent", function(comp, vars)
      vars.perk_gold_is_forever = false
    end)
  end,
  TRICK_BLOOD_MONEY = function(entity_id)
    edit_component2(GameGetWorldStateEntity(), "WorldStateComponent", function(comp, vars)
      vars.perk_trick_kills_blood_money = false
    end)
  end,
  EXPLODING_GOLD = function(entity_id)
    GameRemoveFlagRun("exploding_gold")
  end,
  FASTER_LEVITATION = function(entity_id)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      vars.pixel_gravity = ComponentGetValue2(comp, "pixel_gravity") / 1.4
    end)
  end,
  LOW_GRAVITY = function(entity_id)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      vars.pixel_gravity = ComponentGetValue2(comp, "pixel_gravity") / 0.6
    end)
    local count = get_perk_active_count(entity_id, "$perk_low_gravity")
    -- Remove script and tag if this is the last instance of the perk
    if count == 1 then
      EntityRemoveTag(entity_id, "low_gravity")
      remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/low_gravity.lua" })
    end
  end,
  HIGH_GRAVITY = function(entity_id)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      vars.pixel_gravity = ComponentGetValue2(comp, "pixel_gravity") / 1.4
    end)
    local count = get_perk_active_count(entity_id, "$perk_high_gravity")
    -- Remove script and tag if this is the last instance of the perk
    if count == 1 then
      EntityRemoveTag(entity_id, "high_gravity")
      remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/high_gravity.lua" })
    end
  end,
  BREATH_UNDERWATER = function(entity_id)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      local swim_idle_buoyancy_coeff = {1.2, 0.72, 0.432, 0.2592, 0.15552, 0.093312, 0.0559872, 0.03359232, 0.02015538}
      local swim_up_buoyancy_coeff = {0.9, 0.18, 0.036,0.0072,0.00144,0.000288,5.76e-005,1.152e-005,2.304e-006}
      local swim_down_buoyancy_coeff = {0.7, 0.14, 0.028,0.0056,0.00112,0.000224,4.48e-005,8.96e-006,1.792e-006}
      local swim_drag = {0.95, 1.01,1.01,1.01,1.01,1.01,1.01,1.01,1.01}
      local swim_extra_horizontal_drag = {0.9, 1.01,1.01,1.01,1.01,1.01,1.01,1.01,1.01}
      local count = get_perk_active_count(entity_id, "$perk_breath_underwater")
      count = math.min(9, count)
      vars.swim_idle_buoyancy_coeff = swim_idle_buoyancy_coeff[count]
      vars.swim_up_buoyancy_coeff = swim_up_buoyancy_coeff[count]
      vars.swim_down_buoyancy_coeff = swim_down_buoyancy_coeff[count]
      vars.swim_drag = swim_drag[count]
      vars.swim_extra_horizontal_drag = swim_extra_horizontal_drag[count]
    end)
  end,
  STRONG_KICK = function(entity_id)
    edit_component2(entity_id, "KickComponent", function(comp, vars)
      vars.max_force = ComponentGetValue2(comp, "max_force") / 1.7
      vars.player_kickforce = ComponentGetValue2(comp, "player_kickforce") / 1.7
      vars.kick_damage = ComponentGetValue2(comp, "kick_damage") - 2.4
      vars.kick_knockback = ComponentGetValue2(comp, "kick_knockback") - 250
      vars.telekinesis_throw_speed = ComponentGetValue2(comp, "telekinesis_throw_speed") - 25
      local str = ComponentGetValue2(comp, "kick_entities")
      -- Make a new string up until the last ","
      vars.kick_entities = str:sub(0, str:len() - (str:reverse():find(",") or 0))
    end)
    edit_component2(entity_id, "TelekinesisComponent", function(comp, vars)
      vars.throw_speed = ComponentGetValue2(comp, "throw_speed") - 25
    end)
  end,
  TELEKINESIS = function(entity_id)
    EntityRemoveComponent(entity_id, EntityGetFirstComponentIncludingDisabled(entity_id, "TelekinesisComponent"))
    EntityRemoveComponent(entity_id, EntityGetFirstComponentIncludingDisabled(entity_id, "AudioLoopComponent", "sound_telekinesis_move"))
    EntityRemoveComponent(entity_id, EntityGetFirstComponentIncludingDisabled(entity_id, "AudioLoopComponent", "sound_telekinesis_hold"))
    component_write(EntityGetFirstComponentIncludingDisabled(entity_id, "KickComponent" ), { can_kick = true })
  end,
  EXPLODING_CORPSES = function(entity_id)
    add_halo_level(entity_id, 1)
  end,
  SAVING_GRACE = function(entity_id)
    add_halo_level(entity_id, -1)
  end,
  INVISIBILITY = function(entity_id)
    -- Use default removal function?
    -- Set all player sprites back to alpha 1.0
    local sprite_components = EntityGetComponent(entity_id, "SpriteComponent")
    for i, comp in ipairs(sprite_components or {}) do
      ComponentSetValue2(comp, "alpha", 1.0)
      EntityRefreshSprite(entity_id, comp)
    end
  end,
  GLOBAL_GORE = function(entity_id)
    add_halo_level(entity_id, 1)
  end,
  LEVITATION_TRAIL = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/levitation_trail.lua" })
  end,
  VAMPIRISM = function(entity_id)
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local max_hp = ComponentGetValue2(comp, "max_hp") / 0.75
      vars.max_hp = math.floor(max_hp * 25) / 25
    end)
    add_halo_level(entity_id, 1)
  end,
  EXTRA_HP = function(entity_id)
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local old_max_hp = ComponentGetValue2(comp, "max_hp")
      local current_hp = ComponentGetValue2(comp, "hp")
      vars.max_hp = old_max_hp / 1.5
      current_hp = math.max(current_hp - math.abs(vars.max_hp - old_max_hp), 1 / 25)
      vars.hp = math.min(current_hp, vars.max_hp)
    end)
  end,
  HEARTS_MORE_EXTRA_HP = function(entity_id)
    local var_store = get_variable_storage_component(entity_id, "hearts_more_extra_hp")
    if var_store then
      EntityRemoveComponent(entity_id, var_store)
    end
  end,
  GLASS_CANNON = function(entity_id)
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      vars.hp = 100 / 25
      vars.max_hp = 100 / 25
      vars.max_hp_cap = 0
    end)
  end,
  RESPAWN = function(entity_id)
    add_halo_level(entity_id, -1)
  end,
  FOOD_CLOCK = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/food_clock.lua" })
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "food_clock" })
    local comp = EntityGetFirstComponentIncludingDisabled(entity_id, "IngestionComponent")
    if comp then
      local ingestion_cooldown_delay_frames = {600,400}
      local ingestion_reduce_every_n_frame = {5,3}
      local ingestion_satiation_material_tag = {"", "[food]"}
      local count = get_perk_active_count(entity_id, "$perk_food_clock")
      count = math.min(2, count)
      ComponentSetValue2(comp, "ingestion_cooldown_delay_frames", ingestion_cooldown_delay_frames[count])
      ComponentSetValue2(comp, "ingestion_reduce_every_n_frame", ingestion_reduce_every_n_frame[count])
      ComponentSetValue2(comp, "ingestion_satiation_material_tag", ingestion_satiation_material_tag[count])
    end
  end,
  RADAR_ENEMY = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/radar.lua" })
  end,
  WAND_RADAR = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/radar_wand.lua" })
  end,
  ITEM_RADAR = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/radar_item.lua" })
  end,
  MOON_RADAR = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/radar_moon.lua" })
  end,
  MAP = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/map.xml")
  end,
  TELEPORTITIS_DODGE = function(entity_id)
    for i, child in ipairs(EntityGetAllChildren(entity_id)) do
      if EntityHasTag(child, "teleportitis_dodge") then
        EntityKill(child)
        break
      end
    end
  end,
  NO_WAND_EDITING = function(entity_id)
    edit_component2(GameGetWorldStateEntity(), "WorldStateComponent", function(comp, vars)
      vars.perk_hp_drop_chance = ComponentGetValue2(comp, "perk_hp_drop_chance") - 6
    end)
  end,
  WAND_EXPERIMENTER = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_wand_fired = "data/scripts/perks/wand_experimenter.lua" })
  end,
  ADVENTURER = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/adventurer.lua" })
  end,
  PROJECTILE_HOMING_SHOOTER = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "projectile_homing_shooter" })
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "powerful_shot" })
  end,
  UNLIMITED_SPELLS = function(entity_id)
    edit_component2(GameGetWorldStateEntity(), "WorldStateComponent", function(comp, vars)
      vars.perk_infinite_spells = false
    end)
  end,
  FREEZE_FIELD = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/freeze_field.xml")
  end,
  FIRE_GAS = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/fire_gas.xml")
  end,
  DISSOLVE_POWDERS = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/dissolve_powders.xml")
  end,
  BLEED_SLIME = function(entity_id, perk_entity)
    apply_last_bleed_effect(entity_id, perk_entity)
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local projectile_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "projectile"))
      projectile_resistance = projectile_resistance / 0.75
      ComponentObjectSetValue2(comp, "damage_multipliers", "projectile", projectile_resistance)
    end)
  end,
  BLEED_OIL = function(entity_id, perk_entity)
    apply_last_bleed_effect(entity_id, perk_entity)
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local explosion_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "explosion"))
      explosion_resistance = explosion_resistance / 0.6
      ComponentObjectSetValue2(comp, "damage_multipliers", "explosion", explosion_resistance)
    end)
  end,
  BLEED_GAS = function(entity_id, perk_entity)
    apply_last_bleed_effect(entity_id, perk_entity)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      vars.pixel_gravity = ComponentGetValue2(comp, "pixel_gravity") / 0.75
    end)
  end,
  SHIELD = function(entity_id)
    local shield_num = tonumber(GlobalsGetValue("PERK_SHIELD_COUNT", "0"))
    GlobalsSetValue("PERK_SHIELD_COUNT", tostring(shield_num - 1))
    local last_shield_id
    -- Find the energy shield that was added last and remove it
    for i, child in ipairs(EntityGetAllChildren(entity_id)) do
      if EntityHasTag(child, "energy_shield") then
        last_shield_id = child
      end
    end
    if last_shield_id then
      EntityKill(last_shield_id)
    end
  end,
  REVENGE_EXPLOSION = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_damage_received = "data/scripts/perks/revenge_explosion.lua" })
    edit_all_components2(entity_id, "DamageModelComponent", function(comp, vars)
      local explosion_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "explosion"))
      explosion_resistance = explosion_resistance / 0.75
      ComponentObjectSetValue2(comp, "damage_multipliers", "explosion", explosion_resistance)
    end)
  end,
  REVENGE_TENTACLE = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_damage_received = "data/scripts/perks/revenge_tentacle.lua" })
  end,
  REVENGE_RATS = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_damage_received = "data/scripts/perks/revenge_rats.lua" })
    local new_rattiness subtract_rattiness_level(entity_id)
    if new_rattiness == 2 then
      kill_child_by_name(entity_id, "belt")
      edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
				local vel_x = math.abs(ComponentGetValue2(comp, "velocity_max_x")) / 1.15
				vars.run_velocity = ComponentGetValue2(comp, "run_velocity") / 1.15
				vars.velocity_min_x = -vel_x
				vars.velocity_max_x = vel_x
      end)
    end
    undo_perk_pickup_event("RAT")
  end,
  REVENGE_BULLET = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_damage_received = "data/scripts/perks/revenge_bullet.lua" })
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local explosion_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "explosion"))
      local projectile_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "projectile"))
      explosion_resistance = explosion_resistance / 0.8
      projectile_resistance = projectile_resistance / 0.8
      ComponentObjectSetValue2(comp, "damage_multipliers", "explosion", explosion_resistance)
      ComponentObjectSetValue2(comp, "damage_multipliers", "projectile", projectile_resistance)
    end)
  end,
  ATTACK_FOOT = function(entity_id)
    -- If this was the last perk = remove 4 walkers, remove climber, remove flag run ATTACK_FOOT_CLIMBER
    -- If more than 1 remaining = remove 2 walkers, reduce limb length
    -- If it's the last or second to last one remaining, remove lukkiness level
    -- Always remove 1 attacker
    local children_to_kill = {
      { name = "data/entities/misc/perks/attack_foot/limb_attacker.xml", amount = 1 },
    }
    local remaining_count = get_perk_active_count(entity_id, "$perk_attack_foot")
    if remaining_count == 1 then
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_foot/limb_walker.xml", amount = 4 })
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_foot/limb_climb.xml", amount = 1 })
      GameRemoveFlagRun("ATTACK_FOOT_CLIMBER")
    else
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_foot/limb_walker.xml", amount = 2 })
      -- Reduce limb length
      for _, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
        if EntityHasTag(child, "attack_foot_walker") then
          component_readwrite(EntityGetFirstComponent(child, "IKLimbComponent"), { length = 50 }, function(comp)
            comp.length = comp.length / 1.5
          end)
        end
      end
    end
    if remaining_count < 3 then
      subtract_lukkiness_level(entity_id)
    end
    kill_children_by_filename(entity_id, children_to_kill)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      local run_speed = ComponentGetValue2(comp, "run_velocity") / 1.25
      local vel_x = math.abs(ComponentGetValue2(comp, "velocity_max_x")) / 1.25
      ComponentSetValue2(comp, "run_velocity", run_speed)
      ComponentSetValue2(comp, "velocity_min_x", -vel_x)
      ComponentSetValue2(comp, "velocity_max_x", vel_x)
      if remaining_count == 1 then
        ComponentSetValue2(comp, "pixel_gravity", 350) -- TODO: Remove this once the IKLimbWalkerComponent bug is fixed (hopefully)
      end
    end)
    undo_perk_pickup_event("LUKKI")
  end,
  LEGGY_FEET = function(entity_id)
    local children_to_kill = {
      { name = "data/entities/misc/perks/attack_leggy/leggy_limb_attacker.xml", amount = 1 },
    }
    local remaining_count = get_perk_active_count(entity_id, "$perk_leggy_feet")
    if remaining_count == 1 then
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_leggy/leggy_limb_left.xml", amount = 2 })
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_leggy/leggy_limb_right.xml", amount = 2 })
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_foot/limb_climb.xml", amount = 1 })
      GameRemoveFlagRun("ATTACK_FOOT_CLIMBER")
    else
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_leggy/leggy_limb_left.xml", amount = 1 })
      table.insert(children_to_kill, { name = "data/entities/misc/perks/attack_leggy/leggy_limb_right.xml", amount = 1 })
      -- Reduce limb length
      for _, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
        if EntityHasTag(child, "leggy_foot_walker") then
          component_readwrite(EntityGetFirstComponent(child, "IKLimbComponent"), { length = 50 }, function(comp)
            comp.length = comp.length / 1.5
          end)
        end
      end
    end
    if remaining_count < 3 then
      subtract_lukkiness_level(entity_id)
    end
    kill_children_by_filename(entity_id, children_to_kill)
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      local run_speed = ComponentGetValue2(comp, "run_velocity") / 1.25
      local vel_x = math.abs(ComponentGetValue2(comp, "velocity_max_x")) / 1.25
      ComponentSetValue2(comp, "run_velocity", run_speed)
      ComponentSetValue2(comp, "velocity_min_x", -vel_x)
      ComponentSetValue2(comp, "velocity_max_x", vel_x)
      if remaining_count == 1 then
        ComponentSetValue2(comp, "pixel_gravity", 350) -- TODO: Remove this once the IKLimbWalkerComponent bug is fixed (hopefully)
      end
    end)
    undo_perk_pickup_event("LUKKI")
  end,
  PLAGUE_RATS = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/plague_rats.lua" })
    subtract_rattiness_level(entity_id)
    undo_perk_pickup_event("RAT")
  end,
  VOMIT_RATS = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/vomit_rats.xml")
    subtract_rattiness_level(entity_id)
    undo_perk_pickup_event("RAT")
  end,
  CORDYCEPS = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/cordyceps.lua" })
    GameRemoveFlagRun("player_status_cordyceps")
    subtract_funginess_level(entity_id)
    undo_perk_pickup_event("FUNGI")
  end,
  MOLD = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/slime_fungus.xml")
    GameRemoveFlagRun("player_status_mold")
    subtract_funginess_level(entity_id)
    undo_perk_pickup_event("FUNGI")
  end,
  WORM_SMALLER_HOLES = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/worm_smaller_holes.lua" })
  end,
  PROJECTILE_REPULSION = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/projectile_repulsion_field.xml")
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local projectile_resistance = ComponentObjectGetValue2(comp, "damage_multipliers", "projectile")
      ComponentObjectSetValue2(comp, "damage_multipliers", "projectile", projectile_resistance / 1.26)
    end)
  end,
  RISKY_CRITICAL = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/risky_critical.xml")
  end,
  FUNGAL_DISEASE = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/fungal_disease.xml")
    GameRemoveFlagRun("player_status_fungal_disease")
    subtract_funginess_level(entity_id)
    undo_perk_pickup_event("FUNGI")
  end,
  PROJECTILE_SLOW_FIELD = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/projectile_slow_field.xml")
  end,
  PROJECTILE_REPULSION_SECTOR = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/projectile_repulsion_sector.xml")
  end,
  PROJECTILE_EATER_SECTOR = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/projectile_eater_sector.xml")
  end,
  LOW_RECOIL = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "low_recoil" })
  end,
  ORBIT = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/orbit.lua" })
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local explosion_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "explosion"))
      local projectile_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "projectile"))
      explosion_resistance = explosion_resistance / 1.33
      projectile_resistance = projectile_resistance / 1.33
      ComponentObjectSetValue2(comp, "damage_multipliers", "explosion", explosion_resistance)
      ComponentObjectSetValue2(comp, "damage_multipliers", "projectile", projectile_resistance)
    end)
  end,
  ANGRY_GHOST = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/angry_ghost.xml")
    remove_component(entity_id, "LuaComponent", { script_wand_fired = "data/scripts/perks/angry_ghost_shoot.lua" })
    local remaining_count = get_perk_active_count(entity_id, "$perk_angry_ghost")
    if remaining_count == 1 then
      GameRemoveFlagRun("player_status_angry_ghost")
      subtract_ghostness_level(entity_id)
    end
    undo_perk_pickup_event("GHOST")
  end,
  HUNGRY_GHOST = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/hungry_ghost.xml")
    local remaining_count = get_perk_active_count(entity_id, "$perk_hungry_ghost")
    if remaining_count == 1 then
      GameRemoveFlagRun("player_status_hungry_ghost")
      subtract_ghostness_level(entity_id)
    end
    undo_perk_pickup_event("GHOST")
  end,
  DEATH_GHOST = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/death_ghost.lua" })
    local remaining_count = get_perk_active_count(entity_id, "$perk_death_ghost")
    if remaining_count == 1 then
      GameRemoveFlagRun("player_status_death_ghost")
      subtract_ghostness_level(entity_id)
    end
    undo_perk_pickup_event("GHOST")
  end,
  HOMUNCULUS = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/homunculus_spawner.xml")
    local homunculus = EntityGetWithName("$animal_homunculus")
    if homunculus > 0 then
      EntityKill(homunculus)
    end
  end,
  LUKKI_MINION = function(entity_id)
    local x, y = EntityGetTransform(entity_id)
    local lukkis = EntityGetInRadiusWithTag(x, y, 200, "lukki")
    for i, lukki in ipairs(lukkis) do
      if EntityGetName(lukki) == "$perk_lukki_minion" then
        EntityKill(lukki)
      end
    end
    remove_component(entity_id, "VariableStorageComponent", { name = "owner_id" })
    local remaining_count = get_perk_active_count(entity_id, "$perk_lukki_minion")
    if remaining_count == 1 then
      GameRemoveFlagRun("player_status_lukki_minion")
      subtract_lukkiness_level(entity_id)
    end
    undo_perk_pickup_event("LUKKI")
  end,
  ELECTRICITY = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/electricity.xml")
  end,
  ATTRACT_ITEMS = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/attract_items.lua" })
    local distance_full = tonumber(GlobalsGetValue("PERK_ATTRACT_ITEMS_RANGE", "0"))
    distance_full = distance_full - 24
    if distance_full >= 72 then
      GlobalsSetValue("PERK_ATTRACT_ITEMS_RANGE", tostring(distance_full))
    else
      GlobalsSetValue("PERK_ATTRACT_ITEMS_RANGE", "0")
    end
  end,
  EXTRA_KNOCKBACK = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "extra_knockback" })
  end,
  LOWER_SPREAD = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "lower_spread" })
  end,
  BOUNCE = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "bounce" })
  end,
  FAST_PROJECTILES = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "fast_projectiles" })
  end,
  ALWAYS_CAST = function(entity_id)
    -- In case the currently active item is not a wand, just don't do anything
    pcall(function()
      local wand = EZWand(tonumber(find_the_wand_held(entity_id)))
      local spells, attached_spells = wand:GetSpells()
      wand:DetachSpells()
      for i=1, #attached_spells-1 do
        wand:AttachSpells(attached_spells[i].action_id)
      end
    end)
  end,
  EXTRA_MANA = function(entity_id) end,
  NO_MORE_SHUFFLE = function(entity_id)
    GlobalsSetValue("PERK_NO_MORE_SHUFFLE_WANDS", "0")
  end,
  DUPLICATE_PROJECTILE = function(entity_id)
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "duplicate_projectile" })
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local projectile_resistance = tonumber(ComponentObjectGetValue(comp, "damage_multipliers", "projectile"))
      projectile_resistance = projectile_resistance / 1.25
      ComponentObjectSetValue2(comp, "damage_multipliers", "projectile", projectile_resistance)
    end)
  end,
  FASTER_WANDS = function(entity_id)
    local x, y = EntityGetTransform(entity_id)
    local wands = EntityGetInRadiusWithTag(x, y, 24, "wand")
    for i,wand_id in ipairs(wands) do
      local root_entity = EntityGetRootEntity(wand_id)
      if root_entity == entity_id then
        local wand = EZWand(wand_id)
        wand.rechargeTime = (wand.rechargeTime + 5) / 0.8
        wand.castDelay = (wand.castDelay + 5) / 0.8
        wand.manaChargeSpeed = math.max(10, wand.manaChargeSpeed - 30)
      end
    end
  end,
  EXTRA_SLOTS = function(entity_id)
    local x, y = EntityGetTransform(entity_id)
    local wands = EntityGetInRadiusWithTag(x, y, 24, "wand")
    for i,wand_id in ipairs(wands) do
      local root_entity = EntityGetRootEntity(wand_id)
      if root_entity == entity_id then
        local wand = EZWand(wand_id)
        local old_values = entity_retrieve_table(wand.entity_id, "extra_slots_old_values")
        if old_values then
          wand.capacity = tonumber(old_values.capacity)
        end
      end
    end
  end,
  CONTACT_DAMAGE = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/contact_damage.xml")
  end,
  EXTRA_PERK = function(entity_id)
    local perk_count = tonumber(GlobalsGetValue( "TEMPLE_PERK_COUNT", "3"))
    perk_count = perk_count - 1
    GlobalsSetValue("TEMPLE_PERK_COUNT", tostring(perk_count))
  end,
  PERKS_LOTTERY = function(entity_id)
    local perk_destroy_chance = tonumber(GlobalsGetValue( "TEMPLE_PERK_DESTROY_CHANCE", "100"))
    perk_destroy_chance = perk_destroy_chance * 2
    GlobalsSetValue("TEMPLE_PERK_DESTROY_CHANCE", to_string(perk_destroy_chance))
  end,
  GAMBLE = function(entity_id) end,
  EXTRA_SHOP_ITEM = function(entity_id)
    local shop_item_count = tonumber(GlobalsGetValue("TEMPLE_SHOP_ITEM_COUNT", "5"))
    shop_item_count = math.max(shop_item_count - 1, 5)
    GlobalsSetValue("TEMPLE_SHOP_ITEM_COUNT", tostring(shop_item_count))
  end,
  GENOME_MORE_HATRED = function(entity_id)
    edit_component2(GameGetWorldStateEntity(), "WorldStateComponent", function(comp, vars)
      vars.global_genome_relations_modifier = ComponentGetValue2(comp, "global_genome_relations_modifier") + 25
    end)
    add_halo_level(entity_id, 1)
  end,
  GENOME_MORE_LOVE = function(entity_id)
    edit_component2(GameGetWorldStateEntity(), "WorldStateComponent", function(comp, vars)
      vars.global_genome_relations_modifier = ComponentGetValue2(comp, "global_genome_relations_modifier") - 25
    end)
    add_halo_level(entity_id, -1)
  end,
  PEACE_WITH_GODS = function(entity_id)
    GlobalsSetValue("TEMPLE_PEACE_WITH_GODS", "0")
    local steves = EntityGetWithTag("necromancer_shop")
    for i,steve in ipairs(steves) do
      local game_effect_component = GameGetGameEffect(steve, "CHARM")
      EntityRemoveComponent(steve, game_effect_component)
    end
    add_halo_level(entity_id, -1)
  end,
  MANA_FROM_KILLS = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/mana_from_kills.lua" })
  end,
  ANGRY_LEVITATION = function(entity_id)
    remove_component(entity_id, "LuaComponent", { script_source_file = "data/scripts/perks/angry_levitation.lua" })
  end,
  LASER_AIM = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/laser_aim.xml")
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "laser_aim" })
  end,
  PERSONAL_LASER = function(entity_id)
    kill_child_by_filename(entity_id, "data/entities/misc/perks/personal_laser.xml")
    remove_component(entity_id, "ShotEffectComponent", { extra_modifier = "slow_firing" })
  end,
  MEGA_BEAM_STONE = function(entity_id) end,
}

-- Add the removal functions to the perks table
for i, perk in ipairs(perk_list) do
  if perk_removal_functions[perk.id] then
    perk._remove = perk_removal_functions[perk.id]
  end
end

function remove_component(entity_id, component_name, check_for_vars)
  for i, comp in ipairs(EntityGetComponentIncludingDisabled(entity_id, component_name) or {}) do
    for var, value in pairs(check_for_vars) do
      if ComponentGetValue2(comp, var) == value then
        EntityRemoveComponent(entity_id, comp)
        return
      end
    end
  end
end

function get_perk_active_count(owner_entity, perk_name)
  local count = 0
  for i, child in ipairs(EntityGetAllChildren(owner_entity) or {}) do
    local ui_icon_component = EntityGetFirstComponentIncludingDisabled(child, "UIIconComponent")
    if ComponentGetValue2(ui_icon_component, "name") == perk_name then
      count = count + 1
    end
  end
  return count
end

function subtract_lukkiness_level(entity_id)
	local lukkiness = tonumber(GlobalsGetValue("PLAYER_LUKKINESS_LEVEL", "0"))
	lukkiness = lukkiness - 1
	GlobalsSetValue("PLAYER_LUKKINESS_LEVEL", tostring(lukkiness))
	
	if lukkiness == 2 then
		EntitySetComponentsWithTagEnabled(entity_id, "lukki_enable", false)
		
		local comp = EntityGetFirstComponentIncludingDisabled(entity_id, "SpriteComponent", "lukki_disable")
		if comp then
			ComponentSetValue2(comp, "alpha", 1.0)
		end
		
		local platformingcomponent = EntityGetFirstComponentIncludingDisabled(entity_id, "CharacterPlatformingComponent")
    if platformingcomponent then
      local run_speed = ComponentGetValue2(platformingcomponent, "run_velocity") / 1.1
      local vel_x = math.abs(ComponentGetValue2(platformingcomponent, "velocity_max_x" )) / 1.1
      
      local vel_x_min = 0 - vel_x
      local vel_x_max = vel_x
      
      ComponentSetValue2(platformingcomponent, "run_velocity", run_speed)
      ComponentSetValue2(platformingcomponent, "velocity_min_x", vel_x_min)
      ComponentSetValue2(platformingcomponent, "velocity_max_x", vel_x_max)
    end
	end
end

function subtract_rattiness_level(entity_id)
  local rattiness = tonumber(GlobalsGetValue("PLAYER_RATTINESS_LEVEL", "0"))
  rattiness = rattiness - 1
  GlobalsSetValue("PLAYER_RATTINESS_LEVEL", tostring(rattiness))
  if rattiness == 2 then
    kill_child_by_name(entity_id, "belt")
    edit_component2(entity_id, "CharacterPlatformingComponent", function(comp, vars)
      local vel_x = math.abs(ComponentGetValue2(comp, "velocity_max_x")) / 1.15
      vars.run_velocity = ComponentGetValue2(comp, "run_velocity") / 1.15
      vars.velocity_min_x = -vel_x
      vars.velocity_max_x = vel_x
    end)
  elseif get_rat_perk_count() == 0 then
    local world_entity_id = GameGetWorldStateEntity()
    if world_entity_id then
      component_write(EntityGetFirstComponent(world_entity_id, "WorldStateComponent"), { perk_rats_player_friendly = false })
    end
  end
  return rattiness
end

function subtract_funginess_level(entity_id)
  local funginess = tonumber(GlobalsGetValue("PLAYER_FUNGAL_LEVEL", "0"))
  funginess = funginess - 1
  GlobalsSetValue("PLAYER_FUNGAL_LEVEL", tostring(funginess))
  if funginess == 2 then
		EntitySetComponentsWithTagEnabled(entity_id, "player_hat", false)
		EntitySetComponentsWithTagEnabled(entity_id, "player_hat2_shadow", true)
    edit_component2(entity_id, "DamageModelComponent", function(comp, vars)
      local explosion_resistance = ComponentObjectGetValue2(comp, "damage_multipliers", "explosion")
      ComponentObjectSetValue2(comp, "damage_multipliers", "explosion", explosion_resistance * 0.9)
    end)
  end
  return funginess
end

function subtract_ghostness_level(entity_id)
  local ghostness = tonumber(GlobalsGetValue("PLAYER_GHOSTNESS_LEVEL", "0"))
  ghostness = ghostness - 1
  GlobalsSetValue("PLAYER_GHOSTNESS_LEVEL", tostring(ghostness))
  if ghostness == 2 then
		kill_children_by_filename(entity_id, {
      { name = "data/entities/misc/perks/ghostly_ghost.xml", amount = 1 },
      { name = "data/entities/misc/perks/tiny_ghost_extra.xml", amount = 1 },
    })
    local character_data_component = EntityGetFirstComponentIncludingDisabled(entity_id, "CharacterDataComponent")
    local fly_time = ComponentGetValue2(character_data_component, "fly_recharge_spd") / 1.15
    ComponentSetValue2(character_data_component, "fly_recharge_spd", fly_time)
  end
  return ghostness
end

function get_rat_perk_count(entity_id)
  -- local flag_name = get_perk_picked_flag_name("PLAGUE_RATS")
  -- local pickup_count = tonumber(GlobalsGetValue(flag_name .. "_PICKUP_COUNT", "0"))
  -- local flag_name2 = get_perk_picked_flag_name("REVENGE_RATS")
  -- local pickup_count2 = tonumber(GlobalsGetValue(flag_name2 .. "_PICKUP_COUNT", "0"))
  -- return pickup_count + pickup_count2

  -- TODO: Use the commented out version instead?
  return get_perk_active_count(entity_id, "$perk_plague_rats") + get_perk_active_count(entity_id, "$perk_revenge_rats")
end

function undo_perk_pickup_event(style)
	local flag_name = string.upper(style) .. "_PERK_TOTAL_COUNT"
	local count = tonumber(GlobalsGetValue(flag_name, "0"))
	count = count - 1
	GlobalsSetValue(flag_name, tostring(count))
end

function apply_last_bleed_effect(entity_id, this_perk_entity)
  local active_perks = get_active_perks()
  local bleed_perk_found = false
  -- Iterate backwards to find the last item
  for i=#active_perks,1,-1 do
    local active_perk = active_perks[i]
    if (active_perk.id == "BLEED_OIL" or active_perk.id == "BLEED_SLIME" or active_perk.id == "BLEED_GAS") and active_perk.perk_entity ~= this_perk_entity then
      local perk = get_perk_with_id(perk_list, active_perk.id)
      -- Disable these functions for the duration of the function call
      local _ComponentObjectSetValue = ComponentObjectSetValue
      local _ComponentSetValue2 = ComponentSetValue2
      ComponentObjectSetValue = function() end
      ComponentSetValue2 = function() end
      perk.func(nil, entity_id)
      ComponentObjectSetValue = _ComponentObjectSetValue
      ComponentSetValue2 = _ComponentSetValue2
      bleed_perk_found = true
      break
    end
  end
  -- If no more blood perks are active, return to normal blood
  if not bleed_perk_found then
    local damagemodels = EntityGetComponent(entity_id, "DamageModelComponent")
    if damagemodels then
      for i, damagemodel in ipairs(damagemodels) do
        ComponentSetValue(damagemodel, "blood_material", "blood_fading")
        ComponentSetValue(damagemodel, "blood_spray_material", "blood")
        ComponentSetValue(damagemodel, "blood_multiplier", "1.0")
        ComponentSetValue(damagemodel, "blood_sprite_directional", "")
        ComponentSetValue(damagemodel, "blood_sprite_large", "")
      end
    end
  end
end
