dofile_once("data/scripts/perks/perk.lua")

function string_split(inputstr, sep)
  sep = sep or "%s"
  local t = {}
  for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
    table.insert(t, str)
  end
  return t
end

function entity_store_table(entity_id, var_name, values)
  local var_name = "removable_perks_" .. var_name
  local var_store = get_variable_storage_component(entity_id, var_name)
  local value_string = ""
  for k, v in pairs(values) do
    value_string = value_string .. k .. ";" .. tostring(v) .. ";"
  end
  -- Remove the last semicolon
  value_string = value_string:sub(1, -2)
  if var_store then
    local old_value = ComponentGetValue2(var_store, "value_string")
    ComponentSetValue2(var_store, "value_string", old_value .. ";" .. value_string)
  else
    EntityAddComponent2(entity_id, "VariableStorageComponent", {
      _tags = "enabled_in_hand",
      name = var_name,
      value_string = value_string
    })
  end
end

function entity_retrieve_table(entity_id, var_name, dont_remove_entries)
  local var_name = "removable_perks_" .. var_name
  local var_store = get_variable_storage_component(entity_id, var_name)
  if var_store then
    local out = {}
    local values = string_split(ComponentGetValue2(var_store, "value_string"), ";")
    if #values == 0 then return end
    local num_values = 0
    for i=#values,1,-2 do
      if out[values[i-1]] then
        break
      else
        num_values = num_values + 1
        out[values[i-1]] = values[i]
      end
    end

    if not dont_remove_entries then
      -- Rebuild the string with the last values chopped off
      local new_value_string = ""
      for i=1,#values-num_values*2 do
        new_value_string = new_value_string .. values[i] .. ";"
      end
      -- Remove the last semicolon
      new_value_string = new_value_string:sub(1, -2)
      if new_value_string == "" then
        EntityRemoveComponent(entity_id, var_store)
      else
        ComponentSetValue2(var_store, "value_string", new_value_string)
      end
    end
    return out
  end
end

function kill_child_by_name(entity_id, name)
  for i, child in ipairs(EntityGetAllChildren(entity_id)) do
    if EntityGetName(child) == name then
      EntityKill(child)
      break
    end
  end
end

function is_table_content_equal(t1, t2)
  return table.concat(t1, ",") == table.concat(t2, ",")
end

function is_value_equal(a, b)
  if type(a) ~= type(b) then
    return false
  elseif type(a) == "string" or type(a) == "boolean" then
    return a == b
  end
  return math.abs(a - b) < 0.00001
end

function is_component_value_equal(comp_id, prop, expected_value)
  local prop_val1, prop_val2 = ComponentGetValue2(comp_id, prop)
  if type(expected_value) == "table" then
    if expected_value.min and expected_value.max then
      if not is_value_equal(prop_val1, expected_value.min) or not is_value_equal(prop_val2, expected_value.max) then
        return false
      end
    elseif not is_table_content_equal(prop_val1, expected_value) then
      return false
    end
  elseif not is_value_equal(prop_val1, expected_value) then
    return false
  end
  return true
end

function entity_has_components_with_values(entity, expected_components)
  for _, expected_component in ipairs(expected_components) do
    local comps = EntityGetComponentIncludingDisabled(entity, expected_component.name) or {}
    if #comps == 0 then
      return false
    end
    local found_component = false
    for _, comp_id in ipairs(comps) do
      local match = true
      for prop, expected_value in pairs(expected_component.values or {}) do
        match = match and is_component_value_equal(comp_id, prop, expected_value)
      end
      if match then
        found_component = true
        break
      end
    end
    if not found_component then
      return false
    end
  end
  return true
end

-- BULLSHIT fucking workaround because EntityGetFilename stops working on child entities after a game restart
function entity_is_filename(entity, filename)
  -- If only it was this easy........
  -- do return EntityGetFilename(entity) == filename end
  local lookup = {
    ["data/entities/misc/perks/map.xml"] = "map.lua",
    ["data/entities/particles/perks/projectile_repulsion_field.xml"] = "pe_projectile_repulsion_field.lua",
    ["data/entities/particles/perks/critical_hit_boost.xml"] = "pe_critical_hit_boost.lua",
    ["data/entities/misc/perks/projectile_repulsion_field.xml"] = "projectile_repulsion_field.lua",
    ["data/entities/misc/perks/attack_foot/limb_attacker.xml"] = "attack_foot_limb_attacker.lua",
    ["data/entities/misc/perks/attack_foot/limb_walker.xml"] = "attack_foot_limb_walker.lua",
    ["data/entities/misc/perks/attack_foot/limb_climb.xml"] = "attack_foot_limb_climb.lua",
    ["data/entities/misc/perks/attack_leggy/leggy_limb_attacker.xml"] = "attack_leggy_limb_attacker.lua",
    ["data/entities/misc/perks/attack_leggy/leggy_limb_left.xml"] = "attack_leggy_limb_left.lua",
    ["data/entities/misc/perks/attack_leggy/leggy_limb_right.xml"] = "attack_leggy_limb_right.lua",
    ["data/entities/misc/perks/freeze_field.xml"] = "freeze_field.lua",
    ["data/entities/misc/perks/fire_gas.xml"] = "fire_gas.lua",
    ["data/entities/misc/perks/dissolve_powders.xml"] = "dissolve_powders.lua",
    ["data/entities/misc/perks/vomit_rats.xml"] = "vomit_rats.lua",
    ["data/entities/misc/perks/slime_fungus.xml"] = "slime_fungus.lua",
    ["data/entities/misc/perks/risky_critical.xml"] = "risky_critical.lua",
    ["data/entities/misc/perks/fungal_disease.xml"] = "fungal_disease.lua",
    ["data/entities/misc/perks/projectile_slow_field.xml"] = "projectile_slow_field.lua",
    ["data/entities/particles/perks/projectile_slow_field.xml"] = "pe_projectile_slow_field.lua",
    ["data/entities/misc/perks/projectile_repulsion_sector.xml"] = "projectile_repulsion_sector.lua",
    ["data/entities/misc/perks/projectile_eater_sector.xml"] = "projectile_eater_sector.lua",
    ["data/entities/misc/perks/angry_ghost.xml"] = "angry_ghost.lua",
    ["data/entities/misc/perks/hungry_ghost.xml"] = "hungry_ghost.lua",
    ["data/entities/misc/perks/homunculus_spawner.xml"] = "homunculus_spawner.lua",
    ["data/entities/misc/perks/electricity.xml"] = "electricity.lua",
    ["data/entities/misc/perks/contact_damage.xml"] = "contact_damage.lua",
    ["data/entities/misc/perks/laser_aim.xml"] = "laser_aim.lua",
    ["data/entities/misc/perks/personal_laser.xml"] = "personal_laser.lua",
    ["data/entities/misc/perks/ghostly_ghost.xml"] = "ghostly_ghost.lua",
    ["data/entities/misc/perks/tiny_ghost_extra.xml"] = "tiny_ghost_extra.lua",
    ["data/entities/misc/essences/air.xml"] = "essence_air.lua",
    ["data/entities/misc/essences/fire.xml"] = "essence_fire.lua",
    ["data/entities/misc/essences/alcohol.xml"] = "essence_alcohol.lua",
    ["data/entities/misc/essences/water.xml"] = "essence_water.lua",
    ["data/entities/misc/essences/laser.xml"] = "essence_laser.lua",
  }
  if not lookup[filename] then
    print("<Removable perks> - Missing entity definition for: " .. tostring(filename))
    return false
  end
  return entity_has_components_with_values(entity, dofile_once("mods/removable_perks/files/entity_definitions/" .. lookup[filename]))
end

function kill_child_by_filename(entity_id, filename)
  for i, child in ipairs(EntityGetAllChildren(entity_id)) do
    if entity_is_filename(child, filename) then
      EntityKill(child)
      break
    end
  end
end

function kill_children_by_filename(entity_id, names)
  local already_killed = {}
  for _, child in ipairs(EntityGetAllChildren(entity_id) or {}) do
    for _, name_amount in ipairs(names) do
      if not already_killed[child] and name_amount.amount > 0 and entity_is_filename(child, name_amount.name) then
				name_amount.amount = name_amount.amount - 1
        -- print("Killing " .. tostring(name_amount.name) .. " with id " .. tostring(child))
        already_killed[child] = true
        EntityKill(child)
      end
    end
  end
end

local function remove_perk(perk_, silent, drop_perk)
  local player = EntityGetWithTag("player_unit")[1]
  if not player then
    return
  end
  local perk = get_perk_with_id(perk_list, perk_.id)
  local flag_name = get_perk_picked_flag_name(perk_.id)
	local pickup_count = tonumber(GlobalsGetValue(flag_name .. "_PICKUP_COUNT", "0"))
  pickup_count = pickup_count - 1
	GlobalsSetValue(flag_name .. "_PICKUP_COUNT", tostring(pickup_count))

  -- Remove run flags if this is the last perk instance
  if pickup_count == 0 then
    GameRemoveFlagRun(flag_name)
    for i,other_perk_id in ipairs(perk.remove_other_perks or {}) do
      GameRemoveFlagRun(get_perk_picked_flag_name(other_perk_id))
    end
  end

  local function kill_child_entity_with_game_effect(game_effect)
    if game_effect then
      local game_effect_component = GameGetGameEffect(player, perk.game_effect)
      if game_effect_component then
        for i, child in ipairs(EntityGetAllChildren(player) or {}) do
          for i2, component in ipairs(EntityGetComponentIncludingDisabled(child, "GameEffectComponent") or {}) do
            if component == game_effect_component then
              EntityKill(child)
              return
            end
          end
        end
      end
    end
  end
  kill_child_entity_with_game_effect(perk.game_effect)
  kill_child_entity_with_game_effect(perk.game_effect2)
  if perk.particle_effect then
		kill_child_by_filename(player, ("data/entities/particles/perks/%s.xml"):format(perk.particle_effect))
	end
  if perk._remove then
    perk._remove(player, perk_.perk_entity)
  end
  -- Workaround for a bug, entities without component cannot be EntityKilled
  -- and apparently UIIconComponent doesn't really count as a component
  EntityAddComponent2(perk_.perk_entity, "LifetimeComponent", { lifetime = 1 })
  if not silent then
    GamePrint("Removed perk: " .. GameTextGetTranslatedOrNot(perk.ui_name))
    local x, y = EntityGetTransform(player)
    GamePlaySound("data/audio/Desktop/event_cues.snd", "event_cues/chest/create", x, y)
  end
  if drop_perk then
    local x, y = EntityGetTransform(player)
    local perk_entity = perk_spawn(x, y - 5, perk_.id)
    EntityAddComponent2(perk_entity, "VariableStorageComponent", {
      name = "perk_dont_remove_others",
      value_bool = true
    })
  end
end

local function remove_essence(essence, silent, drop_essence)
  local player = EntityGetWithTag("player_unit")[1]
  if not player then
    return
  end
  if essence.id == "fire" then
    local damage_model_component = EntityGetFirstComponentIncludingDisabled(player, "DamageModelComponent")
    local projectile_resistance = ComponentObjectGetValue2(damage_model_component, "damage_multipliers", "projectile")
    local explosion_resistance = ComponentObjectGetValue2(damage_model_component, "damage_multipliers", "explosion")
    ComponentObjectSetValue2(damage_model_component, "damage_multipliers", "projectile", projectile_resistance / 1.3)
    ComponentObjectSetValue2(damage_model_component, "damage_multipliers", "explosion", explosion_resistance / 1.3)
  end

  kill_child_by_filename(player, "data/entities/misc/essences/" .. essence.id .. ".xml")

  GameRemoveFlagRun("essence_" .. essence.id)
	local globalskey = "ESSENCE_" .. string.upper(essence.id) .. "_PICKUP_COUNT"
	local pickups = tonumber(GlobalsGetValue(globalskey, "0"))
	pickups = pickups - 1
	GlobalsSetValue(globalskey, tostring(pickups))

  EntityKill(essence.perk_entity)
  if not silent then
    GamePrint("Removed essence: " .. GameTextGetTranslatedOrNot(essence.ui_name))
    local x, y = EntityGetTransform(player)
    GamePlaySound("data/audio/Desktop/event_cues.snd", "event_cues/chest/create", x, y)
  end
  if drop_essence then
    local x, y = EntityGetTransform(player)
    EntityLoad("data/entities/items/pickup/essence_" .. essence.id .. ".xml", x, y - 5)
  end
end

function is_gamepad_active()
	local players = EntityGetWithTag("player_unit")
	for i, player in ipairs(players) do
		local platform_shooter_player = EntityGetFirstComponentIncludingDisabled(player, "PlatformShooterPlayerComponent")
		if platform_shooter_player then
			local is_gamepad = ComponentGetValue2(platform_shooter_player, "mHasGamepadControlsPrev")
			if is_gamepad then
        return true
			end
		end
	end
end

return {
  remove_perk = remove_perk,
  remove_essence = remove_essence,
}
